/* This file is part of the Zenipex Library (zenilib).
 * Copyleft (C) 2011 Mitchell Keith Bloch (bazald).
 *
 * This source file is simply under the public domain.
 */

#include <zenilib.h>

#if defined(_DEBUG) && defined(_WINDOWS)
#define DEBUG_NEW new(_NORMAL_BLOCK, __FILE__, __LINE__)
#define new DEBUG_NEW
#endif

using namespace std;
using namespace Zeni;

static std::vector<String> g_args;

class Model_Viewer_State : public Gamestate_Base {
public:
  Model_Viewer_State()
    : m_model("models/MODEL.3DS"),
    m_projector(make_pair(Point2f(0.0f, 0.0f), Point2f(100.0f, float(get_Window().get_height()))),
                make_pair(Point2i(0, 0), Point2i(100, get_Window().get_height()))),
    m_ambient(Point2f(10.0f, 40.0f), Point2f(90.0f, 40.0f), 15.0f),
    m_diffuse(Point2f(10.0f, 100.0f), Point2f(90.0f, 100.0f), 15.0f),
    m_specular(Point2f(10.0f, 160.0f), Point2f(90.0f, 160.0f), 15.0f),
    m_pos_x(Point2f(10.0f, 220.0f), Point2f(90.0f, 220.0f), 15.0f),
    m_pos_y(Point2f(10.0f, 280.0f), Point2f(90.0f, 280.0f), 15.0f),
    m_rotation(Point2f(110.0f, get_Window().get_height() - 50.0f),
               Point2f(get_Window().get_width() - 10.0f, get_Window().get_height() - 50.0f),
               15.0f, 0.0f),
    m_keyframe(Point2f(110.0f, get_Window().get_height() - 20.0f),
               Point2f(get_Window().get_width() - 10.0f, get_Window().get_height() - 20.0f),
               15.0f, 0.0f),
    m_prev_clear_color(get_Video().get_clear_Color())
  {
    const Widget_Renderer_Slider wrs(get_Colors()["slider_line"], get_Colors()["slider"]);
    m_ambient.fax_Renderer(&wrs);
    m_diffuse.fax_Renderer(&wrs);
    m_specular.fax_Renderer(&wrs);
    m_pos_x.fax_Renderer(&wrs);
    m_pos_y.fax_Renderer(&wrs);
    m_rotation.fax_Renderer(&wrs);
    m_keyframe.fax_Renderer(&wrs);

    set_pausable(true);

    // Scale the Model to fit in a box of edge length 200.0f
    const Model_Extents me = m_model.get_extents();
    const float model_diameter = (me.upper_bound - me.lower_bound).magnitude();
    const float model_scale = 200.0f / model_diameter;
    m_model.set_scale(Vector3f(model_scale, model_scale, model_scale));

    // Center the Model in world space.
    m_model.set_translate(-Vector3f(m_model.get_position()));

    // Set up the Camera
    m_camera.near_clip = 100.0f;
    Projector3D projector(m_camera);
    const float camera_radius = projector.unproject(Vector3f()).magnitude();
    m_camera.position = Point3f(0.0f, camera_radius, 0.0f);
    m_camera.look_at(Point3f());
    m_camera.near_clip = 1.0f;
    m_camera.far_clip = 2.0f * camera_radius;

    // Misc.
    get_Video().set_clear_Color(get_Colors()["bg_color"]);

    m_sliders.lend_Widget(m_ambient);
    m_sliders.lend_Widget(m_diffuse);
    m_sliders.lend_Widget(m_specular);
    m_sliders.lend_Widget(m_pos_x);
    m_sliders.lend_Widget(m_pos_y);
    m_sliders.lend_Widget(m_rotation);
    m_sliders.lend_Widget(m_keyframe);
  }

  ~Model_Viewer_State() {
    get_Video().set_clear_Color(m_prev_clear_color);
  }

private:
  void on_mouse_motion(const SDL_MouseMotionEvent &event) {
    m_sliders.on_event(event, m_projector);
  }

  void on_mouse_button(const SDL_MouseButtonEvent &event) {
    m_sliders.on_event(event, m_projector);
  }

  void perform_logic() {
    /** Handle Controls **/

    m_light.ambient = Color(1.0f,
      m_ambient.get_slider_position(),
      m_ambient.get_slider_position(),
      m_ambient.get_slider_position());
    m_light.diffuse = Color(1.0f,
      m_diffuse.get_slider_position(),
      m_diffuse.get_slider_position(),
      m_diffuse.get_slider_position());
    m_light.specular = Color(1.0f,
      m_specular.get_slider_position(),
      m_specular.get_slider_position(),
      m_specular.get_slider_position());

    const Model_Extents me = m_model.get_extents();

    m_light.position = Point3f(
      100.0f * (m_pos_x.get_slider_position() - 0.5f) * (me.upper_bound.x - me.lower_bound.x),
      100.0f * (m_pos_y.get_slider_position() - 0.5f) * (me.upper_bound.y - me.lower_bound.y),
      100.0f * (me.upper_bound.y - me.lower_bound.y));

    m_model.set_rotate(m_rotation.get_slider_position() * 2.0f * Zeni::Global::pi + Zeni::Global::pi, Vector3f(0.0f, 0.0f, 1.0f));

    m_model.set_keyframe(m_model.get_keyframes() * m_keyframe.get_slider_position());
  }

  void render() {
    Video &vr = get_Video();
    Font &font = get_Fonts()["font"];
    Colors &colors = get_Colors();

    /** Render Model **/

    vr.set_3d_view(m_camera,
      make_pair(Point2i(100, 0), Point2i(get_Window().get_width(), get_Window().get_height())));

    vr.set_lighting(true);
    vr.set_ambient_lighting(Color(1.0f, 0.0f, 0.0f, 0.0f));
    vr.set_Light(0, m_light);

    m_model.render();

    /** Render Controls **/

    //vr.set_2d_view(
    //  make_pair(Point2f(0.0f, 0.0f), Point2f(100.0f, float(Video::get_height()))),
    //  make_pair(Point2i(0, 0), Point2i(100, Video::get_height())));
    vr.set_2d();

    vr.set_lighting(false);

    font.render_text("Ambient", Point2f(10.0f, 10.0f), colors["slider_text"]);
    font.render_text("Diffuse", Point2f(10.0f, 70.0f), colors["slider_text"]);
    font.render_text("Specular", Point2f(10.0f, 130.0f), colors["slider_text"]);
    font.render_text("X Position", Point2f(10.0f, 190.0f), colors["slider_text"]);
    font.render_text("Y Position", Point2f(10.0f, 250.0f), colors["slider_text"]);
    font.render_text("Rotate", Point2f(10.0f, float(get_Window().get_height() - 60.0f)), colors["slider_text"]);
    font.render_text("Keyframe", Point2f(10.0f, float(get_Window().get_height() - 30.0f)), colors["slider_text"]);

    m_sliders.render();
  }

  /** Model **/

  Model m_model;
  Camera m_camera;

  float min_distance;

  /** Controls **/

  Projector2D m_projector;

  Light m_light;

  Slider m_ambient;
  Slider m_diffuse;
  Slider m_specular;

  Slider m_pos_x;
  Slider m_pos_y;

  Slider m_rotation;
  Slider m_keyframe;

  Widgets m_sliders;

  Color m_prev_clear_color;
};

class Bootstrap {
  class Gamestate_One_Initializer : public Gamestate_Zero_Initializer {
    virtual Gamestate_Base * operator()() {
      Window::set_title("zenilib Model Viewer");

      get_Controllers();
      get_Video();
      get_Colors().load_file("config/mv_colors.xml");
      get_Textures().load_file("config/mv_textures.xml");
      get_Fonts().load_file("config/mv_fonts.xml");
      get_Sounds();

      return new Model_Viewer_State;
    }
  } m_goi;

public:
  Bootstrap() {
    g_gzi = &m_goi;
  }
} g_bootstrap;

int main(int argc, char **argv) {
  g_args.reserve(argc);
  for(const char * const *at = argv, * const *aend = argv + argc; at != aend; ++at)
    g_args.push_back(*at);

  return zenilib_main(argc, argv);
}
